# Module of TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2014 Wave Systems Corp.
# Copyright (C) 2014 Peter Thoeny, peter[at]thoeny.org 
# Copyright (C) 2014 TWiki Contributors. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.

=pod

---+ package TWiki::LoginManager::EmailTwoStepAuth

=cut

package TWiki::LoginManager::EmailTwoStepAuth;

use strict;
use Assert;
use Error qw( :try );

our $sessionVarName = '_EmailTwoStepAuthAccessCode';

=pod

---++ ClassMethod new ($session, $impl)

Construct the two-step auth management object

=cut

sub new {
    my ( $class, $twiki ) = @_;
    my $this = bless( { twiki => $twiki }, $class );

    return $this;
}

=begin twiki

---++ ObjectMethod finish()

Break circular references.

=cut

# Note to developers; please undef *all* fields in the object explicitly,
# whether they are references or not. That way this method is "golden
# documentation" of the live fields in the object.
sub finish {
    my $this = shift;
    undef $this->{twiki};
}

=pod

---++ ObjectMethod secondStepAuth()

Do second step authentication:

   * Send user an e-mail with one time access code
   * Show dialog to enter access code

=cut

sub secondStepAuth {
    my( $this, $loginName, $origUrl ) = @_;
    my $session = $this->{twiki};

    # skip second auth if in command line context
    return '' if $session->inContext( 'command_line' );

    # check whitelist
    my $addr = $session->{request}->remoteAddress() || '';
    my $whitelist = $TWiki::cfg{EmailTwoStepAuthContrib}{WhitelistAddresses} || '';
    $whitelist =~ s/[^0-9\.\,]//g;
    $whitelist =~ s/,/\|/g;
    if( $addr && $whitelist && $addr =~ /^($whitelist)/ ) {
        # trusted environment, skip second step authentication
        return '';
    }

    # two-step authentication is required - generate one-time-use access code
    my $accessCode = 'ac' . sprintf( '%02d', rand(100) ) . '-' . sprintf( '%04d', rand(10000) );
    my $wikiName = $session->{users}->getWikiName( $loginName );
    my @emails = map{ my $a = "$wikiName <$_>"; $a; } $session->{users}->getEmails( $loginName );
    unless( scalar @emails ) {
        # non found, so get e-mails from Email form field of user profile topic
        my @users = ( $wikiName );
        my $result = TWiki::Func::searchInWebContent(
          '%META:FIELD{name="Email"',
          $TWiki::cfg{UsersWebName},
          \@users, {} );
        foreach my $user (keys %$result ) {
            foreach my $line ( @{$result->{$user}} ) {
                if( $line =~ s/^.*value="([^"]*)".*$/$1/o ) {
                    push( @emails, "$user <$line>" );
                }
            }
        }
    }

    # save access code for later verification
    my $now = time();
    TWiki::Func::setSessionValue( $sessionVarName, "$accessCode:$now:$loginName" );

    # send e-mail to login user with access code
    my $template = $TWiki::cfg{EmailTwoStepAuthContrib}{MessageTmpl} || 'emailtwostepmessage';
    my $tmpl = $session->templates->readTemplate( $template, $session->getSkin() );
    return "Two-step authentication installation error: $template template not found" unless( $tmpl );
    $tmpl =~ s/%EMAILADDRESS%/join( ', ', @emails )/geo;
    $tmpl =~ s/%ACCESSCODE%/$accessCode/go;
    $tmpl = $session->handleCommonTags( $tmpl, $session->{webName}, $session->{topicName} );
    $tmpl =~ s/<nop>//g;
    my $warnings = $session->net->sendEmail( $tmpl );
    return "$warnings <hr /><pre>$tmpl</pre>" if( $warnings );

    # load and return "enter access code" template
    $template = $TWiki::cfg{EmailTwoStepAuthContrib}{LoginTmpl} || 'emailtwosteplogin';
    $tmpl = $session->templates->readTemplate( $template, $session->getSkin() )
            || "Two-step authentication installation error: $template template not found";
    $tmpl =~ s/%LOGINNAME%/$loginName/go;
    $tmpl =~ s/%ORIGURL%/$origUrl/go;
    return $tmpl;
}

=pod

---++ ObjectMethod verifyAuth()

Verify access code on second step authentication. Return empty string if OK, else return error string.

=cut

sub verifyAuth {
    my( $this, $loginName, $accessCode ) = @_;
    my $session = $this->{twiki};

    return '' if $session->inContext( 'command_line' );

    # compare to saved access code
    my ( $expectedAC, $timestamp, $expectedLN )
      = split( /:/, TWiki::Func::getSessionValue( $sessionVarName ), 3 );
    TWiki::Func::clearSessionValue( $sessionVarName ); # clear session variable (one time use only)
    my $maxAge = $TWiki::cfg{EmailTwoStepAuthContrib}{MaxAge} || 600;
    my $error = '';
    unless( $accessCode
            && ( $accessCode eq $expectedAC )
            && ( $loginName eq $expectedLN ) 
            && $timestamp + $maxAge > time()
    ) {
        $error = $TWiki::cfg{EmailTwoStepAuthContrib}{AcessCodeError} ||
                 'Invalid or outdated access code, please try again';
    }
    return $error;
}

1;
